"""
    CUSTOM READERS CLASSES
        - Class which manages reader tasks like auth, requests, pagination
"""
import os

from sailthru import SailthruClient

from sdc_dp_helpers.api_utilities.date_managers import date_handler, date_range
from sdc_dp_helpers.api_utilities.retry_managers import request_handler
from sdc_dp_helpers.sailthru.config_magagers import get_config


class CustomSailThruReader:
    """
        Custom SailThru Reader
    """

    def __init__(self, **kwargs):
        self.config = get_config(kwargs.get('config_path', None))
        self.credentials = get_config(kwargs.get('credentials_path', None))

        self.sailthru_client = SailthruClient(
            api_key=self.credentials.get('api_key', None),
            secret=self.credentials.get('api_secret', None)
        )

    def get_templates(self):
        """
        Dynamically gathers all templates for 'send' stat.
        """
        response = self.sailthru_client.get_templates()
        if response.is_ok():
            return response.get_body()

        raise EnvironmentError(
            f'Error: {response.get_error().get_error_code()}\n'
            f'Mesage: {response.get_error().get_message()}'
        )

    @request_handler(
        wait=int(os.environ.get('REQUEST_WAIT_TIME', 0)),
        backoff_factor=float(os.environ.get('REQUEST_BACKOFF_FACTOR', 0.01)),
        backoff_method=os.environ.get('REQUEST_BACKOFF_METHOD', 0.01)
    )
    def request(self, data, action='stats'):
        """
        Request various stats from Sailthru about primary list membership
        or campaign and triggered message activity.
        Endpoint URL: https://api.sailthru.com/stats
        Additional parameters are dependent on the stat value
        the type of stats you want to request:
            - list
            - blast
            - send
        """
        print(f'Action: {action}\nQuery: {data}')

        response = self.sailthru_client.api_get(
            action=action, data=data
        )

        if not response.is_ok():
            error = response.get_error()
            message = 'APIError:{}, Status Code:{}, Error Code:{}'.format(
                error.get_message(),
                str(response.get_status_code()),
                str(error.get_error_code()))
            if (
                    str(response.get_status_code()) != '404' and
                    str(error.get_error_code()) != '99'
            ):
                raise RuntimeError(message)

            # If there is no data for something like a template,
            # just a warning is printed.
            print(f'Warning: No data in given instance.\n {message}')
        elif response.is_ok():
            return response.get_body()
        else:
            return None

    def run_query(self):
        """
        Consumes a .yaml config file and loops through the date and url
        to return relevant data from Sailthru Get Request.
        """
        data_set = []
        for stat, data in self.config.items():
            data['stat'] = stat
            date_range_list = date_range(
                start_date=data['start_date'],
                end_date=data['end_date']
            )

            # send requires template ids for the request,
            # so all possible templates are generated and loop through
            if stat == 'send':
                templates = self.get_templates()
                # we need to split send date ranges for day-by-day values
                for list_date in date_range_list:
                    for template in templates.get('templates'):
                        template_name = template.get('name')

                        print(f'Template: {template_name}')
                        data['template'] = template_name

                        data['start_date'] = list_date
                        data['end_date'] = list_date
                        request_data = self.request(data=data)

                        if request_data is not None:
                            request_data['stat'] = 'send'
                            request_data['template'] = template_name
                            data_set.append(request_data)

            # blast requires the action be converted from 'stat' to 'blast'
            elif stat == 'blast':
                data['start_date'] = date_handler(data['start_date'])
                data['end_date'] = date_handler(data['end_date'])
                request_data = self.request(action=stat, data=data)
                if request_data is not None:
                    request_data['stat'] = 'blast'
                    data_set.append(request_data)

            elif stat == 'list':
                # list stat does not support start to end date scope, only a single 'date'
                for list_date in date_range_list:
                    print(f'List date range at: {list_date}.')
                    data['date'] = list_date
                    request_data = self.request(data=data)
                    if request_data is not None:
                        request_data['stat'] = data['stat']
                        data_set.append(request_data)

            else:
                raise ValueError(f'Specified stat {stat} is not supported.')

        return data_set



"""
    CUSTOM READERS CLASSES
        - Class which manages reader tasks like auth, requests, pagination
"""
import json

from azure.storage.blob import ContainerClient
from azure.core.exceptions import ResourceNotFoundError


class CustomAzureReader():
    """
        Custom Azure Reader
    """

    # pylint: disable=super-init-not-called
    def __init__(self, sas_token_path: str):
        self.container_client = self.get_container_client(sas_token_path)

    @staticmethod
    def get_container_client(
            sas_token_path
    ):
        """
        Create a client to communicate with Sailthru
        Args:
            sas_token_path: The file path to sas token
        Returns:
            A service that is connected to the specified API.
        """
        with open(sas_token_path) as json_file:
            data = json.load(json_file)

        return ContainerClient.from_container_url(data['sas_token'])

    def pull_file(self, file_name):
        """
            Pulls the data file from Azure blob
            :return: storage downloader object
        """
        try:
            byte_sequence = self.container_client.download_blob(file_name).readall()
            return json.loads(byte_sequence.decode("utf-8"))
        except ResourceNotFoundError as exception:
            print(
                'Could not find specified blob: {exception}'.format(
                    exception=str(exception)
                )
            )

            return None




import gzip
import json
from urllib.parse import urlparse

import boto3
from pyspark.sql import functions as f
from pyspark.sql import types as t


def convert(field):
    """
    Handles nested json strings that need to retain their structure.
    This also guarantees that data is not lost during schema inference.
    """
    if not isinstance(field.dataType, t.StructType) and not isinstance(field.dataType, t.ArrayType):
        return field.name
    return f.to_json(field.name).alias(field.name)


def read_gz(file_url):
    """
    Reads gzipped metadata json file that has no .gz extension for spark to identify.
    """
    client = boto3.client('s3')

    url = urlparse(file_url)
    bucket, filepath = url.netloc, url.path.lstrip("/")

    obj = client.get_object(
        Bucket=bucket,
        Key=filepath
    )

    gzfile = gzip.decompress(obj['Body'].read())
    return json.loads(gzfile)


def filter_paths(bucket, prefix, position):
    """
    Gathers relevant paths in an s3 bucket and groups them by
    the specified position in the url.
    This can be used to read and write whilst retaining a specified
    level of file structuring of objects in buckets.

    Prefix can be used to speed up the search process.

    If the S3 urls are:
        s3://dl-base/metric-1/date-1
        s3://dl-base/metric-2/date-1
        s3://dl-base/metric-2/date-2
        s3://dl-base/metric-3/date-1

    One could set position to 4 and return:
        {
            'metric-1': ['s3://dl-base/metric-1/date-1'],
            'metric-2': ['s3://dl-base/metric-2/date-1', 's3://dl-base/metric-2/date-2'],
            'metric-3': ['s3://dl-base/metric-3/date-1']
        }

    """
    client = boto3.client('s3')
    paginator = client.get_paginator('list_objects')
    page_iterator = paginator.paginate(Bucket=bucket, Prefix=prefix)

    paths = []
    for page in page_iterator:
        keys = [key['Key'] for key in page['Contents']]
        for key in keys:
            key_path = f's3://{bucket}/{key}'
            if len(key_path.split('.')) > 1:
                paths.append(key_path)

    paths_dict = {}  # get all data paths associated with position
    for path in paths:
        identifier = path.split('/')[position]
        if identifier not in paths_dict:
            paths_dict[identifier] = []
        paths_dict[identifier].append(path)

    return paths_dict



# pylint: disable=no-self-use
import json
import time
from datetime import datetime
from urllib.error import URLError

import pandas as pd
import requests

from sdc_dp_helpers.api_utilities.date_managers import date_handler
from sdc_dp_helpers.api_utilities.file_managers import load_file
from sdc_dp_helpers.api_utilities.retry_managers import retry_handler


class CustomOneSignalReader:
    def __init__(self, creds_file, config_file=None):
        self._creds = load_file(creds_file, "yml")
        self._config = load_file(config_file, "yml")

        self._header = {
            "Content-Type": "application/json; charset=utf-8",
            "Authorization": f"Basic {self._creds.get('api_key')}",
            "User-Agent": "Mozilla/5.0",
        }

        self._request_session = requests.Session()

    @retry_handler(exceptions=KeyError, total_tries=5, should_raise=True)
    def _csv_export_query(self):
        """
        Query handler for csv_export.
        https://documentation.onesignal.com/reference/csv-export

        Generate a compressed CSV export of all of your current user data.
        This method can be used to generate a compressed CSV export of all
        of your current user data. It is a much faster alternative than
        retrieving this data using the /players API endpoint.
        The file will be compressed using GZip.
        The file may take several minutes to generate depending on the number
        of users in your app.
        The URL generated will be available for 3 days and includes random v4
        uuid as part of the resource name to be unguessable.

        ⚠ Note that adding any date oriented payload drastically affects
        the output data. Even if the payload values are null.
        """
        print("POST: csv_export.")
        # response is not always successful for some reason, so retry handler
        # handles the missing "csv_file_url" key response
        response: dict = self._request_session.post(
            url="https://onesignal.com/api/v1/players/csv_export",
            headers=self._header,
            json={"app_id": self._creds.get("app_id")},
        ).json()

        # status code is not always returned, so handling error strings instead
        errors = response.get("errors", [None])[0]
        if errors:
            # The header is definitely valid, so we must assume the app_id is not
            if (
                errors
                == "app_id not found. You may be missing a Content-Type: application/json header."
            ):
                raise EnvironmentError(
                    "App Id may not be valid or removed from the account."
                )

            # Seems like we are restricted to making one query at a time for a given user, so wait 5 minutes
            if (
                errors
                == "User already running another CSV export. Please wait until your other CSV exporter finishes and try again."
            ):
                time.sleep(300)  # the reader may try again after the 5 minute wait
                raise EnvironmentError(errors)

            # Raise all other unknown errors in a standard fashion
            raise EnvironmentError(errors)
        else:
            print("No errors raised, fetching url.")

        response = response["csv_file_url"]
        attempts = 0
        while True:
            try:
                # Making use of pandas to stream the
                # compressed csv data directly to a df
                _date_fmt = "%Y-%m-%d"
                data_frame: pd.DataFrame = pd.read_csv(response)

                # added functionality to filter results by created_at
                # note the filter can not exceed the last 30 days
                sd = date_handler(self._config.get("start_date", None), _date_fmt)
                ed = date_handler(self._config.get("end_date", None), _date_fmt)
                if sd is not None and ed is not None:
                    # using a temp date field to select relevant data before dropping
                    data_frame["tmp_date"] = pd.to_datetime(data_frame["created_at"])
                    data_frame["tmp_date"] = data_frame["tmp_date"].dt.strftime(
                        _date_fmt
                    )
                    data_frame = data_frame[
                        (
                            data_frame["tmp_date"]
                            >= datetime.strptime(sd, _date_fmt).strftime(_date_fmt)
                        )
                        & (
                            data_frame["tmp_date"]
                            <= datetime.strptime(ed, _date_fmt).strftime(_date_fmt)
                        )
                    ]
                    data_frame.drop("tmp_date", axis="columns", inplace=True)
                break
            except URLError:
                # max wait time is 15 minutes at 10 attempts before we need to contact Onesignal
                print(f"Waiting for file to generate, attempt {attempts}.")
                time.sleep(90)

            attempts += 1

            if attempts >= 10:
                raise URLError(
                    f"Csv file could not be generated, contact Onesignal support. "
                    f"Response {response}."
                )

        return data_frame.to_json(orient="records")

    def _view_notifications_query(self):
        """
        Query handler for view_notifications.
        https://documentation.onesignal.com/reference/view-notifications

        View the details of multiple notifications.
        OneSignal periodically deletes records of API notifications
        older than 30 days.
        If you would like to export all notification data to CSV,
        you can do this through the dashboard.

        ⚠ Note that adding any date oriented payload drastically affects
        the output data. Even if the payload values are null.
        """
        print("GET: view_notifications.")
        initial_response = self._request_session.get(
            url="https://onesignal.com/api/v1/notifications",
            headers=self._header,
            json={"app_id": self._creds.get("app_id")},
        )

        total_offset = initial_response.json().get("total_count")

        data_set = []
        for offset in range(0, total_offset):
            print(f"At offset: {offset} of {total_offset}.")
            response = self._request_session.get(
                url="https://onesignal.com/api/v1/notifications",
                headers=self._header,
                json={"app_id": self._creds.get("app_id"), "offset": offset},
            )

            if response.status_code == 200:
                # if no more data is present skip response
                data = response.json().get("notifications", None)
                if data is None:
                    continue
                data_set.append(data)
            else:
                raise requests.exceptions.ConnectionError(
                    f"Connection failed with {response.status_code} and {response.reason}."
                )

        return json.dumps(data_set)

    def run_query(self):
        """
        The Onesignal API provides programmatic methods to
        access data in Onesignal for view_notifications and csv_exports.
        """
        _endpoint = self._config.get("endpoint")
        if _endpoint == "csv_export":
            return self._csv_export_query()
        elif _endpoint == "view_notifications":
            return self._view_notifications_query()
        else:
            raise EnvironmentError(f"Endpoint {_endpoint} is not supported.")



import json

import yaml


def load_file(file_location, fmt):
    """
        Gathers file data from json or yaml.
    """
    if file_location is not None:
        if fmt in ['yaml', 'yml']:
            with open(file_location, 'r') as file:
                config = yaml.safe_load(file)
        elif fmt == 'json':
            with open(file_location, 'r') as file:
                config = json.load(file)

    return config



"""
    Set up the sdc_helpers package
"""
from distutils.command.sdist import sdist as _sdist

from setuptools import setup


class sdistzip(_sdist):
    """Override sdist to always produce .zip archive"""

    def initialize_options(self):
        _sdist.initialize_options(self)
        self.formats = 'zip'


setup(
    name='sdc_dp_helpers',
    packages=
    [
        'sdc_dp_helpers',
        'sdc_dp_helpers.google_analytics',
        'sdc_dp_helpers.google_search_console',
        'sdc_dp_helpers.sailthru',
        'sdc_dp_helpers.azure',
        'sdc_dp_helpers.api_utilities',
        'sdc_dp_helpers.falcon',
        'sdc_dp_helpers.pyspark',
        'sdc_dp_helpers.onesignal',
        'sdc_dp_helpers.zoho',
        'sdc_dp_helpers.google_ads'
    ],
    install_requires=[
        'boto3',
        'google-api-python-client',
        'httplib2',
        'pathlib',
        'oauth2client',
        'numpy',
        'pandas',
        'pyOpenSSL',
        'python-interface',
        'sailthru-client',
        'azure-storage-blob',
        'sdc-helpers==1.6.2',
        'requests',
        'zcrmsdk==2.0.10',
        'mysql',
        'mysql-connector',
        'httplib2',
        'googleads==28.0.0',
        'google-ads==13.0.0',
        'protobuf==3.17.3'
    ],
    extras_require={'pyspark': ['pyspark']},
    cmdclass={'sdist_zip': sdistzip},
    description="A module for developing data pipelines from external api\'s and on ETL like services",
    version='1.2.8',
    url='http://github.com/RingierIMU/sdc-dataPipeline-helpers',
    author='Ringier South Africa',
    author_email='tools@ringier.co.za',
    keywords=[
        'pip',
        'datapipeline',
        'helpers',
    ],
    download_url='https://github.com/RingierIMU/sdc-global-dataPipeline-helpers/archive/v0.9.0.zip'
)



"""
    utils functions
"""
import yaml


def init_config(file_name: str = './example.yaml'):
    """
        Util method that creates an example config at the given file_name
        args:
            file_name (str): full path to config YAML file
    """
    config = {
        'dateRanges': [{
            'startDate': '2020-02-01',
            'endDate': '2020-02-01'}],
        'samplingLevel': 'LARGE',
        'pageSize': 100,
        'chunk_date_range': True,
        'dayFreq': 1,
        'viewIDs': ['<add client viewID here>'],
        'queryConfig': [{
            'metrics': [
                {'expression': 'ga:sessions',
                 'alias': 'sessions'}],
            'dimensions': [
                {'name': 'ga:date'},
                {'name': 'ga:medium'},
                {'name': 'ga:source'},
                {'name': 'ga:keyword'},
                {'name': 'ga:hostname'},
                {'name': 'ga:deviceCategory'}],
            'orderBys': [
                {'fieldName': 'ga:date',
                 'sortOrder': 'DESCENDING'}]}]}

    # write to file
    with open(file_name, 'w') as file_:
        yaml.dump(config, file_)



"""
    Retry Managers:
        - retry decorators
        - failure mode handlers
"""
import time
from functools import wraps
from random import random

from sdc_helpers.slack_helper import SlackHelper


def retry_handler(
        exceptions,
        total_tries: int = 4,
        initial_wait: float = 0.5,
        backoff_factor: int = 2,
        should_raise: bool = False
):
    """
        Decorator - managing API failures

        args:
            exceptions (Exception): Exception instance or list of
            Exception instances to catch & retry
            total_tries (int): Total retry attempts
            initial_wait (float): initial delay between retry attempts in seconds
            backoff_factor (int): multiplier used to further randomize back off
            should_raise (bool): Raise exception after all retires fail

        return:
            wrapped function's response
    """

    def retry_decorator(function):
        @wraps(function)
        def func_with_retries(*args, **kwargs):
            """wrapper function to decorate function with retry functionality"""
            _tries, _delay = total_tries, initial_wait
            while _tries > 0:
                try:
                    print(f'Attempt {total_tries + 1 - _tries}')
                    return function(*args, **kwargs)
                except exceptions as exception:
                    # get logger message
                    if _tries == 1:
                        slack_helper = SlackHelper()
                        slack_helper.send_critical(
                            message=str(
                                f'Function: {function.__name__}\n'
                                f'Failed despite best efforts after {total_tries} tries\n'
                                f'Exception {exception}.'
                            )
                        )

                        if should_raise:
                            raise exception
                    else:
                        msg = str(f'Function: {function.__name__}\n'
                                  f'Exception {exception}.\n'
                                  f'Retrying in {_delay} seconds!')
                    # log with print
                    print(msg)
                    # decrement _tries
                    _tries -= 1
                    # pause
                    time.sleep(_delay)
                    # increase delay by backoff factor
                    _delay = _delay * backoff_factor

        return func_with_retries

    return retry_decorator


def request_handler(
        wait: float = 0,
        backoff_factor: float = 0.01,
        backoff_method: str = 'random'
):
    """
        Decorator - managing API failures

        Args:
            wait (float, optional): [description].
                Defaults to 100 seconds for GoogleAnalytics.
                Initial delay between retry attempts in seconds
            backoff_factor (float, optional):
                [description]. Defaults to 0.1.
                Multiplier used to further randomize delay
            backoff_method (str, optional):
                [description]. Defaults to 'random'. hHow to apply backoff
        return:
            wrapped function's response
    """

    def retry_decorator(function):
        @wraps(function)
        def func_with_retries(*args, **kwargs):
            """wrapper function to decorate function with retry functionality"""
            if backoff_method == 'random':
                _delay = wait * (1 + backoff_factor * random())
            else:
                _delay = wait

            print(f'Waiting {_delay} seconds before attempt')
            time.sleep(_delay)
            return function(*args, **kwargs)

        return func_with_retries

    return retry_decorator
